################################################################
#
# Name:          Followfy
# Version:       0.2 (2014 05 15)
# 
# Description:   Follow me with solidsolving behaviour for planar paths
# 
# Author:        Anders Lyhagen 
#
# Usage:         Followfy under the Plugins menu
#                
#                Select an open path and a profile placed at either end. Run Followfy. 
#                If the profile is attached to the path a triple click selection is
#                admitted.
################################################################

module CAUL_Followfy

  ##################################
  ####### EXTRACT SELECTION ########
  ##################################
  
  #Selections admitted:
  # 1) Edge path + face
  # 2) Edge path + face with edges (triple click)
  def self.extractSelection(sel)
    
    fs = sel.grep(Sketchup::Face)
    es = sel.grep(Sketchup::Edge)
    
    #validate content of selection
    return [nil, nil] if (fs.length != 1 || es.length == 0) 
    
    fh = extractFaceLoopHash(fs.first) 
    sh = extractEdgeHash(es, fh)
    ps = extractPath(sh)
    
    return ps != nil ? [ps, fs.first] : [nil, nil]
  end
  
  #extract a hash with all edges connected to the face (outer
  #and inner loops)
  def self.extractFaceLoopHash(f)
    fh = {}
    f.loops.each { |l| l.edges.each { |e| fh[e] = nil }}
    return fh
  end
  
  #extract a hash with all edges in the array es, except those in hash fh
  def self.extractEdgeHash(es, fh)
    sh = {}
    es.each { |e| sh[e] = nil unless fh.has_key?(e) }
    return sh
  end
  
  ##############################
  ######### EXTRACT PATH #######
  ##############################
    
  #extract a path (ordered array of positions) from the edges in hash sh
  def self.extractPath(sh)  
    
    ps = extractEndPoints(sh)
       
    if ps.empty? #loop
      e = sh.shift[0] #extract random start edge...
      sh[e] = nil #....and put it back
      vs = [e.start]
    elsif ps.length == 2 #open path
      e = ps.first[0]
      vs = [ps.first[1], getOtherVertex(e, ps.first[1])]
    else #non valid path
      return nil
    end
    
    #extract the path
    trace(vs, sh, e, vs.last)
    
    #validate the path
    if ps.empty? && vs.length == sh.length #valid loop
      vs << e.start #add first vertex to tail as well
      return getPositionPath(vs)
    elsif ps.length == 2 && vs.length == (sh.length + 1) #valid open path
      return getPositionPath(vs)
    else
      return nil
    end
  end
 
  #return an array of [[e, v], ..] representing the endpoints of the path
  #(the method does some double work with vertices)
  def self.extractEndPoints(sh)
    ps = []
    sh.each_key { |e| 
      ps << [e, e.start] if endPoint?(e, e.start, sh)   
      ps << [e, e.end]   if endPoint?(e, e.end, sh)
    }
    return ps
  end
  
  #v is an end vertex if e (connected to v) is the only edge in sh
  #connected to v
  def self.endPoint?(e, v, sh)
    return v.edges.all? { |e2| !sh.has_key?(e2) || e == e2 }
  end
  
  def self.trace(vs, sh, e, v)
    eh = {e => nil}
    until (e, v = nextEdge(sh, eh, e, v)) == [nil, nil]
      eh[e] = nil
      vs << v
    end
  end
  
  #return the next edge (connected to e with v) and its other vertex.
  #The next edge must be present in sh and excluded from eh
  def self.nextEdge(sh, eh, e, v)
    v.edges.each { |ne|
      next if e == ne
      return [nil, nil] if eh.has_key?(ne) #loop
      return [ne,  getOtherVertex(ne, v)] if sh.has_key?(ne) 
    }
    return nil, nil
  end
  
  def self.getOtherVertex(e, v)
    return e.start == v ? e.end : e.start
  end
   
  def self.getPositionPath(vs)
    return vs.map { |v| v.position }
  end
   
  #########################
  ######## OFFSET #########
  #########################
    
  def self.getProfileFromFace(f)
    return f.outer_loop.vertices.map { |v| v.position }
  end

  #modify ps in order to fit with the profile (reverse, rotate, truncate, etc)
  def self.relatePathAndProfile(ps, prf)
     #just check start and end (for the time being)
     ps.reverse! if ps.first.distance(prf.first) > ps.last.distance(prf.first)
  end

  def self.getPathVectors(ps)
    vps = []
    for i in 0..ps.length - 2
      vps[i] = (ps[i + 1] - ps[i]).normalize!
    end
    
    if ps.first != ps.last #open path
      vps[i + 1] = vps.last
    elsif
      puts 'GetPathVectors: loop case not implemented'
    end    
    return vps
  end
  
  def self.getCornerPlanes(ps, vs)
    cps = [[ps.first, vs.first]]
    for i in 1..ps.length - 1
      cps[i] = [ps[i], (vs[i - 1] + vs[i]).normalize!]
    end
    
    if (ps.first != ps.last) #open path
      cps[i] = [ps.last, vs.last]
    elsif #closed path
      puts 'getCornerPlanes:  loop case not implemented'
    end
    return cps
  end
  
  #FollowMe
  def self.getOffsetPaths(prf, cps, vs)
    return prf.map { |p| getOffsetPath(p, cps, vs) }
  end
  
  def self.getOffsetPath(p, cps, vs)
    op = [Geom::intersect_line_plane([p, vs.first], cps.first)] #project onto first plane
    for i in 1..cps.length - 1
      op[i] = Geom::intersect_line_plane([op[i - 1], vs[i - 1]], cps[i])
    end
    return op
  end
  
  #########################
  ##### MODIFY PATHS ######
  #########################
  
  #This part is basically a hack...

  #remove loops and gaps from offsetpaths
  def self.modifyOffsetPaths(ent, ps, ops, vps)
    m = getRetroPaths(ops, vps)
    m.each { |op| 
      removeLoops(ent, ps, op) 
      removeLoops(ent, ps, op) #make second pass for added robustness  TODO: fix this..
    }
   
    #we assume that the rest of the retro motions correspond to gaps
    m = getRetroPaths(ops, vps)    
    #turn gaps into loops...
    m.each { |op| 
      inds = getRetroIndeces(op, vps)
      inds.each { |i| 
        len = op[i].distance(op[i + 1])
        a0 = vps[i - 1].angle_between(vps[i])
        a1 = vps[i].angle_between(vps[i + 1])
        
        #extend a gap into a loop
        if a0 > a1
          v = vps[i - 1].clone
          v.length = len
          op[i].offset! v
        elsif
          v = vps[i + 1].clone
          v.length = -len
          op[i + 1].offset! v
        end
      }
    }
    
    #remove new loops produced from gaps...
    m = getRetroPaths(ops, vps)
    m.each { |op| 
      removeLoops(ent, ps, op) 
      removeLoops(ent, ps, op) #make second pass for added robustness...
    }
  end
  
  def self.getRetroIndeces(op, vps)
    ri = []
    for i in 0..op.length - 2
      ri << i if vps[i].dot(op[i + 1] - op[i]) < 0
    end
    return ri
  end
  
  #filter out the offset paths that need modification
  def self.getRetroPaths(ops, vps)
    m = [];
    ops.each { |op| m << op if retroMotion?(op, vps) }
    return m
  end
  
  #a problematic offset path must have a retro motion, ie. a segments that goes in the
  #opposite direction of the corresponding path segment
  def self.retroMotion?(op, vps)
    for i in 0..op.length - 2
      return true if vps[i].dot(op[i + 1] - op[i]) < 0
    end
    return false
  end
  
  def self.dummyExplode(ent, g0)
    ng = ent.add_group
    ng.entities.add_instance(g0.entities.parent, g0.transformation)
    g0.erase!
    ng.entities[0].explode
    return ng
  end
  
  def self.removeLoops(ent, ps, op)
    g = ent.add_group
    for i in 0..op.length - 2
      g.entities.add_line(op[i], op[i + 1])
    end
    g = dummyExplode(ent, g)
    
    loop_count = 0
    vs = g.entities.grep(Sketchup::Edge).map{|e| e.vertices}.flatten!.uniq!
    vs.each { |v|
      if v.edges.length > 2 #fork
        fps = getForkPositions(v)
        mm = getForkIndex(fps, op)
        modifyLoop(mm, op, v.position)
        loop_count += 1
      end
    }
    g.erase!
    #puts 'RemoveLoops: '+'Loops: '+loop_count.to_s
  end
  
  #mm contains the indeces before and after the fork in op. fp is the position of the fork
  def self.modifyLoop(mm, op, fp)
    for i in mm[0] + 1..mm[1] - 1
      op[i] = fp
    end
  end
  
  def self.getForkPositions(v) 
    ps = []
    v.edges.each { |e| ps << getOtherVertex(e, v).position unless e.deleted? }
    return ps
  end
  
  #return highest and lowest index in ps related to the fork
  def self.getForkIndex(fps, op)
    min = 1000000; max = -1000000;
    fps.each { |fp|
      for i in 0..op.length
        if(op[i] == fp)
          min = i if i < min
          max = i if i > max
        end
      end
    }
    return [min, max]
  end

  ########################
  ####### SKIN ###########
  ########################
  
  def self.skin(g, ops)
    ent = g.entities
    for j in 0..ops[0].length - 2 #along path
      for i in 0..ops.length - 1  #around profile
        ni = (i + 1) % ops.length  
        skinQuad(ent, ops[i][j] , ops[i][j + 1], 
                      ops[ni][j], ops[ni][j + 1])
      end
    end
    
    #plug the ends (we need to check for loop here...)
    ent.add_face getProfileFace(ops, 0)
    ent.add_face getProfileFace(ops, ops[0].length - 1)
    
    mergeCoPlanarFaces(g)
  end
  
  def self.getProfileFace(ops, ind)
    prf = []
    for i in 0..ops.length - 1
      prf << ops[i][ind]
    end
    return prf
  end
    
  def self.skinQuad(ent, p00, p01, p10, p11)
    ent.add_face(p00, p01, p11) unless duplicate?(p00, p01, p11)
    ent.add_face(p10, p00, p11) unless duplicate?(p10, p00, p11)
  end
  
  def self.duplicate?(p0, p1, p2)
    return p0 == p1 || p0 == p2 || p1 == p2
  end

  def self.mergeCoPlanarFaces(g)
    #filter out potential coplanar faces
    es = g.entities.grep(Sketchup::Edge).select{|e| e.faces.length == 2 && 
                                                    e.faces[0].normal.dot(e.faces[1].normal) > 0.99990}
    es.each { |e|
        vs = e.faces[0].vertices + e.faces[1].vertices
        p = Geom::fit_plane_to_points vs
        cop = vs.all? { |v| v.position.on_plane? p }
        e.erase! if cop
    }
  end
  
  #########################
  ######## MAIN ###########
  #########################
  
  def self.debug(ent, ps, vps, cps, ops)
    ng = ent.add_group
    ops.each { |op|
      next if op == nil
      for i in 0..op.length - 2
        opv = (op[i + 1] - op[i])
          ng.entities.add_line op[i], op[i + 1]
      end
   }
  end
  
  def self.followfy(ent, sel)
    
    ps, f = extractSelection(sel)
    return if ps == nil
    
    ng = ent.add_group #create result group
    
    prf = getProfileFromFace(f)
    relatePathAndProfile(ps, prf) 
    vps = getPathVectors(ps)
    cps = getCornerPlanes(ps, vps)
    ops = getOffsetPaths(prf, cps, vps)
    modifyOffsetPaths(ent, ps, ops, vps)    
    skin(ng, ops)
    #debug(ent, ps, vps, cps, ops)
  end

  def self.followfyMain(mode)
    mod = Sketchup.active_model
    ent = mod.entities 
    sel = mod.selection
    
    mod.start_operation("followfy", true)
    followfy(ent, sel)
    mod.commit_operation
  end
  
  unless file_loaded?( __FILE__ )
    menu = UI.menu( 'Plugins' )
    menu.add_item( 'Followfy' ) { self.followfyMain(0)}
  end

  file_loaded( __FILE__ )

  followfyMain(0)
end


 

